

import os
import collections


FileSystemStats = collections.namedtuple("FileSystemStats", [
	"bytesTotal", "bytesUsed", "bytesFreeSystem", "bytesFreeUser", "rateBytesUsed",
	"inodesTotal", "inodesUsed", "inodesFree", "rateInodesUsed"
])

def getFileSystemStats(mountPointPath):
	statvfs = os.statvfs(mountPointPath)

	bytesTotal = statvfs.f_frsize * statvfs.f_blocks		# Size of filesystem in bytes
	bytesFreeSystem = statvfs.f_frsize * statvfs.f_bfree	# Actual number of free bytes
	bytesFreeUser = statvfs.f_frsize * statvfs.f_bavail		# Number of free bytes that ordinary users are allowed to use (excl. reserved space)
	bytesUsed = bytesTotal - bytesFreeSystem
	rateBytesUsed = ((bytesTotal - bytesFreeUser) / bytesTotal) if bytesTotal > 0 else 1

	inodesTotal = statvfs.f_files    					    # inodes
	inodesFree = statvfs.f_ffree	   						# free inodes
	inodesUsed = inodesTotal - inodesFree
	rateInodesUsed = (inodesUsed / inodesTotal) if inodesTotal > 0 else 1

	return FileSystemStats(
		bytesTotal, bytesUsed, bytesFreeSystem, bytesFreeUser, rateBytesUsed,
		inodesTotal, inodesUsed, inodesFree, rateInodesUsed
		)
#

def findMountPoint(path:str):
    path = os.path.abspath(path)
    orig_dev = os.stat(path).st_dev

    while path != '/':
        pdir = os.path.dirname(path)
        if os.stat(pdir).st_dev != orig_dev:
            # we crossed the device border
            break
        path = pdir
    return path
#

def getFolderSize(dirPath:str):
    total_size = os.path.getsize(dirPath)
    for item in os.listdir(dirPath):
        itempath = os.path.join(dirPath, item)
        if os.path.isfile(itempath):
            total_size += os.path.getsize(itempath)
        elif os.path.isdir(itempath):
            total_size += getFolderSize(itempath)
    return total_size
#




#
# This code is copied from: https://gist.github.com/mkropat/7559409
# Author: mkropat
# Liense: MIT
#



import os

class UserNotFound(Exception): pass

### Windows ###
if os.name == 'nt':
    import ctypes
    from ctypes import windll, wintypes

    class WinError:     # [1]
        InvalidParameter = 87
        InvalidSid = 1337

    class NtStatus:     # [2]
        Success         = 0x00000000
        NoneMapped      = 0xC0000073

    class AccessMask:   # [3]
        Delete                      = 0x00010000
        ReadControl                 = 0x00020000
        WriteDAC                    = 0x00040000
        WriteOwner                  = 0x00080000
        Synchronize                 = 0x00100000
        StandardRightsRequired      = 0x000F0000
        StandardRightsRead          = ReadControl
        StandardRightsWrite         = ReadControl
        StandardRightsExecute       = ReadControl
        StandardRightsAll           = 0x001F0000
        SpecificRightsAll           = 0x0000FFFF

    class TokenAccessLevels:        # See [4] and TokenAccessLevels.cs in Mono source
        AssignPrimary       = 1<<0
        Duplicate           = 1<<1
        Impersonate         = 1<<2
        Query               = 1<<3
        QuerySource         = 1<<4
        AdjustPrivileges    = 1<<5
        AdjustGroups        = 1<<6
        AdjustDefault       = 1<<7
        AdjustSessionId     = 1<<8
        Read                = AccessMask.StandardRightsRead | Query
        Write               = AccessMask.StandardRightsWrite | AdjustPrivileges | AdjustGroups | AdjustDefault
        AllAccess           = AccessMask.StandardRightsRequired | 0x1FF
        MaximumAllowed      = 0x2000000

    class TokenInformationClass:    # [5]
        User,\
        Groups,\
        Privileges,\
        Owner,\
        PrimaryGroup,\
        DefaultDacl,\
        Source,\
        Type,\
        ImpersonationLevel,\
        Statistics,\
        RestrictedSids,\
        SessionId,\
        GroupsAndPrivileges,\
        SessionReference,\
        SandBoxInert,\
        AuditPolicy,\
        Origin,\
        ElevationType,\
        LinkedToken,\
        Elevation,\
        HasRestrictions,\
        AccessInformation,\
        VirtualizationAllowed,\
        VirtualizationEnabled,\
        IntegrityLevel,\
        UIAccess,\
        MandatoryPolicy,\
        LogonSid,\
        IsAppContainer,\
        Capabilities,\
        AppContainerSid,\
        AppContainerNumber,\
        UserClaimAttributes,\
        DeviceClaimAttributes,\
        RestrictedUserClaimAttributes,\
        RestrictedDeviceClaimAttributes,\
        DeviceGroups,\
        RestrictedDeviceGroups,\
        SecurityAttributes,\
        IsRestricted,\
        = range(1, 41)

    class SidNameUse:   # [6]
        TypeUser,\
        TypeGroup,\
        TypeDomain,\
        TypeAlias,\
        TypeWellKnownGroup,\
        TypeDeletedAccount,\
        TypeInvalid,\
        TypeUnknown,\
        TypeComputer,\
        TypeLabel,\
        = range(1, 11)

    # Type declarations
    pSID = wintypes.HANDLE
    ENUM = ctypes.c_uint

    class SidAndAttributes(ctypes.Structure):   # [7]
        _fields_ = (
            ('Sid', pSID),
            ('Attributes', wintypes.DWORD),
        )

    class TokenUser(ctypes.Structure):          # [8]
        _fields_ = (
            ('User', SidAndAttributes),
        )

    _GetCurrentProcess = windll.kernel32.GetCurrentProcess      # [9]
    _GetCurrentProcess.restype = wintypes.HANDLE

    _OpenProcessToken = windll.advapi32.OpenProcessToken        # [10]
    _OpenProcessToken.restype = wintypes.BOOL
    _OpenProcessToken.argtypes = (
        wintypes.HANDLE,                # ProcessHandle
        wintypes.DWORD,                 # DesiredAccess
        ctypes.POINTER(wintypes.HANDLE) # TokenHandle (out)
    )

    _GetTokenInformation = windll.advapi32.GetTokenInformation  # [11]
    _GetTokenInformation.restype = wintypes.BOOL
    _GetTokenInformation.argtypes = (
        wintypes.HANDLE,    # TokenHandle
        ENUM,               # TokenInformationClass
        ctypes.c_void_p,    # TokenInformation (out)
        wintypes.DWORD,     # TokenInformationLength
        wintypes.PDWORD     # ReturnLength (out)
    )

    _ConvertStringSidToSid = windll.advapi32.ConvertStringSidToSidW     # [12]
    _ConvertStringSidToSid.restype = wintypes.BOOL
    _ConvertStringSidToSid.argtypes = (
        wintypes.LPWSTR,        # StringSid
        ctypes.POINTER(pSID),   # Sid (out)
    )

    _CovertSidToStringSid = windll.advapi32.ConvertSidToStringSidW      # [13]
    _CovertSidToStringSid.restype = wintypes.BOOL
    _CovertSidToStringSid.argtypes = (
        pSID,                           # Sid
        ctypes.POINTER(wintypes.LPWSTR) # StringSid (out)
    )

    _LocalFree = windll.kernel32.LocalFree      # [14]
    _LocalFree.restype = None
    _LocalFree.argtypes = (wintypes.HANDLE,)

    _IsValidSid = windll.advapi32.IsValidSid    # [15]
    _IsValidSid.restype = wintypes.BOOL
    _IsValidSid.argtypes = (pSID,)

    _LookupAccountSid = windll.advapi32.LookupAccountSidW   # [16]
    _LookupAccountSid.restype = wintypes.BOOL
    _LookupAccountSid.argtypes = (
        wintypes.LPWSTR,        # lpSystemName (optional)
        pSID,                   # lpSid
        wintypes.LPWSTR,        # lpName (out, optional)
        wintypes.LPDWORD,       # cchName (in/out)
        wintypes.LPWSTR,        # lpReferencedDomainName (out, optional)
        wintypes.LPDWORD,       # cchReferencedDomainName (in/out)
        ctypes.POINTER(ENUM),   # peUse (out)
    )

    class Sid:
        def __init__(self, sid):
            if isinstance(sid, str):
                self.pointer = pSID()
                status = _ConvertStringSidToSid(sid, self.pointer)
                if status == WinError.InvalidSid:
                    raise ValueError('Not a valid sid: %s' % sid)
                elif not status:
                    raise ctypes.WinError()
            else:
                if not _IsValidSid(sid):
                    raise ValueError('Invalid sid: %s' % uid)
                self.pointer = sid

        def __str__(self):
            buf = wintypes.LPWSTR()
            if not _CovertSidToStringSid(self.pointer, buf):
                raise ctypes.WinError()
            string_sid = buf.value
            _LocalFree(buf)
            return string_sid

    def _getpsid():
        proc_token = wintypes.HANDLE()
        if not _OpenProcessToken(_GetCurrentProcess(), TokenAccessLevels.Read, proc_token):
            raise ctypes.WinError()

        # Get required length by calling with a 0 length buffer
        length = wintypes.DWORD()
        _GetTokenInformation(proc_token, TokenInformationClass.User, None, 0, length)

        buf = ctypes.create_string_buffer(length.value)
        if not _GetTokenInformation(proc_token, TokenInformationClass.User, buf, len(buf), length):
            raise ctypes.WinError()

        token_user = ctypes.cast(buf, ctypes.POINTER(TokenUser)).contents
        return token_user.User.Sid


    def getuid():
        return str(Sid(_getpsid()))

    def lookup_username(uid=None):
        sid = Sid(uid or _getpsid())

        use = ENUM()
        cchName = wintypes.DWORD()
        cchReferencedDomainName = wintypes.DWORD()

        # Get required lengths by calling with 0 length buffers
        _LookupAccountSid(None, sid.pointer, None, cchName, None, cchReferencedDomainName, use)

        name = ctypes.create_unicode_buffer(cchName.value)
        domain = ctypes.create_unicode_buffer(cchReferencedDomainName.value)

        status = _LookupAccountSid(None, sid.pointer, name, cchName, domain, cchReferencedDomainName, use)
        if status == NtStatus.NoneMapped:
            raise UserNotFound('No such sid: %s' % sid)
        elif not status:
            raise ctypes.WinError()

        if use.value != SidNameUse.TypeUser:
            raise ValueError('Not a user sid: %s' % sid)

        return name.value

### Posix ###
else:
    import pwd
    from os import getuid

    def lookup_username(uid=None):
        if uid is None: uid = getuid()
        try:
            return pwd.getpwuid(uid)[0]
        except (KeyError, TypeError):
            raise UserNotFound('No such uid: %s' % uid)


__all__ = (getuid, lookup_username)

if __name__ == '__main__':
    print('%s %s' % (getuid(), lookup_username()))

# References:
# [1] http://msdn.microsoft.com/en-us/library/windows/desktop/ms681381
# [2] http://msdn.microsoft.com/en-us/library/cc704588
# [3] http://msdn.microsoft.com/en-us/library/windows/desktop/aa374892
# [4] http://msdn.microsoft.com/en-us/library/aa374905
# [5] http://msdn.microsoft.com/en-us/library/windows/desktop/aa379626
# [6] http://msdn.microsoft.com/en-us/library/windows/desktop/aa379601
# [7] http://msdn.microsoft.com/en-us/library/windows/desktop/aa379595
# [8] http://msdn.microsoft.com/en-us/library/windows/desktop/aa379634
# [9] http://msdn.microsoft.com/en-us/library/windows/desktop/ms683179
# [10] http://msdn.microsoft.com/en-us/library/aa379295
# [11] http://msdn.microsoft.com/en-us/library/windows/desktop/aa446671
# [12] http://msdn.microsoft.com/en-us/library/aa376402
# [13] http://msdn.microsoft.com/en-us/library/windows/desktop/aa376399
# [14] http://msdn.microsoft.com/en-us/library/windows/desktop/aa366730
# [15] http://msdn.microsoft.com/en-us/library/windows/desktop/aa379151
# [16] http://msdn.microsoft.com/library/windows/desktop/aa379166





import os
import sys
import netifaces





#
# Retrieve all MAC addresses from all active network adapters (excluding local host)
#
def getMACs():
	ret = []
	for i in netifaces.interfaces():
		addrs = netifaces.ifaddresses(i)
		try:
			if_mac = addrs[netifaces.AF_LINK][0]['addr']
			if_ip = addrs[netifaces.AF_INET][0]['addr']
			if if_ip == "127.0.0.1":
				continue
			ret.append(if_mac)
		except IndexError as e:
			pass
		except KeyError as e:
			pass
	return ret










import os
import sys
import netifaces





#
# Retrieve all IP addresses from all active network adapters (excluding local host)
#
def getIPs():
	ret = []
	for i in netifaces.interfaces():
		addrs = netifaces.ifaddresses(i)
		try:
			if_mac = addrs[netifaces.AF_LINK][0]['addr']
			if_ip = addrs[netifaces.AF_INET][0]['addr']
			if if_ip == "127.0.0.1":
				continue
			ret.append(if_ip)
		except IndexError as e:
			pass
		except KeyError as e:
			pass
	return ret
#




class LocalIPAddressDetector(object):

	def __init__(self, localIPAddressFilters:list):
		self.__localIPAddressRegExes = []
		for s in localIPAddressFilters:
			pos = s.index("*")
			if pos < 0:
				sRegex = "^" + s + "$"
			elif pos == 0:
				sRegex = "^.+" + s + "$"
			elif pos == len(s) - 1:
				sRegex = "^" + s + ".+$"
			else:
				raise Exception("Invalid network card pattern: " + repr(s))
			self.__localIPAddressRegExes.append(re.compile(sRegex))
	#

	def __call__(self):
		return self.getIPAddresses()
	#

	#
	# Retrieve all local IP addresses this service might be concated in a local network.
	#
	def getIPAddresses(self):
		ret = []

		ipAddrTuples = []
		for ifaceName in netifaces.interfaces():
			addrs = netifaces.ifaddresses(ifaceName)
			try:
				#if_mac = addrs[netifaces.AF_LINK][0]['addr']
				ipAddr = addrs[netifaces.AF_INET][0]['addr']
				if ipAddr == "127.0.0.1":
					continue
				ipAddrTuples.append([ifaceName, ipAddr, False])
			except IndexError as e:
				pass
			except KeyError as e:
				pass

		for ipAddrRegEx in self.__localIPAddressRegExes:
			for r in ipAddrTuples:
				if r[2]:
					continue
				if ipAddrRegEx.match(r[0]):
					r[2] = True
					ret.append(r[1])
		
		return ret
	#

#









import typing
import os
import subprocess
import pwd
import signal





#
# Kill a process.
#
# @param	int|dict process			A processes, either as <c>int</c> (= process ID) or as <c>dict</c>
#										as provided by <c>jk_sysinfo.get_ps()</c>.
# @param	AbstractLogger log			A logger that will receive log messages.
# @return	bool						Returns <c>true</c> on success. <c>false</c> is returned if something went wrong.
#										(In that case error log messages might have been generated, but no detailed error
#										information is returned.)
#
def killProcess(process:typing.Union[int,dict], log) -> bool:
	assert isinstance(process, (int,dict))

	# ----

	if process is None:
		if log:
			log.notice("No processes to kill.")
		return True

	ret = True

	try:
		if isinstance(process, int):
			pid = process
		else:
			assert isinstance(process, dict)
			pid = process["pid"]
			assert isinstance(pid, int)

		if log:
			log.notice("Killing: " + str(pid))
		os.kill(pid, signal.SIGTERM)
	except Exception as ee:
		if log:
			log.error(ee)
		ret = False

	return ret
#



#
# Kill a set of processes (in the order they are presented).
#
# @param	int[]|dict[] processes		A list of processes, either as <c>int[]</c> of process IDs or as <c>dict[]</c>
#										as provided by <c>jk_sysinfo.get_ps()</c>.
# @param	AbstractLogger log			A logger that will receive log messages.
# @return	bool						Returns <c>true</c> on success. <c>false</c> is returned if something went wrong.
#										(In that case error log messages might have been generated, but no detailed error
#										information is returned.)
#
def killProcesses(processes:typing.Union[list, tuple], log) -> bool:
	assert isinstance(processes, (list, tuple))

	# ----

	if (processes is None) or (not processes):
		if log:
			log.notice("No processes to kill.")
		return True

	ret = True

	for process in processes:
		try:
			if isinstance(process, int):
				pid = process
			else:
				assert isinstance(process, dict)
				pid = process["pid"]
				assert isinstance(pid, int)

			if log:
				log.notice("Killing: " + str(pid))
			os.kill(pid, signal.SIGTERM)
		except Exception as ee:
			if log:
				log.error(ee)
			ret = False

	return ret
#



def _demote(user_uid, user_gid):

	def fn():
		os.setgid(user_gid)
		os.setuid(user_uid)
	#

	return fn

#

#
# Run a process as other user and wait until it terminates. No input or output is processed.
# Of course for this to work you should be user <c>root</c>.
#
# @param	str accountName			The name of the user account under which to execute the process.
# @param	str filePath			The path of the file to execute. This can be a bash script.
# @param	str[] arguments			(optional) A list of arguments for the program to execute.
# @param	AbstractLogger log		A logger that will receive log messages.
# @return	bool					Returns <c>true</c> on success. <c>false</c> is returned if something went wrong.
#									(In that case error log messages might have been generated, but no detailed error
#									information is returned.)
#
def runProcessAsOtherUser(accountName:str, filePath:str, args:typing.Union[list,tuple], log) -> bool:
	assert isinstance(accountName, str)
	assert accountName

	assert isinstance(filePath, str)
	assert filePath
	assert os.path.isfile(filePath)

	if args is None:
		args = []
	else:
		if isinstance(args, tuple):
			args = list(args)
		else:
			assert isinstance(args, list)

	assert log

	# ----

	if log:
		log.notice("Running " + repr(filePath) + " as " + repr(accountName) + "...")

	scriptDirPath = os.path.dirname(filePath)

	pw_record = pwd.getpwnam(accountName)
	accountName = pw_record.pw_name
	user_home_dir = pw_record.pw_dir
	uid = pw_record.pw_uid
	gid = pw_record.pw_gid

	currentUserID = os.getuid()
	if currentUserID != 0:
		if uid != currentUserID:
			raise Exception("Must be root to run processes as other user!")

	env = os.environ.copy()
	env["HOME"     ]  = user_home_dir
	env["LOGNAME"  ]  = accountName
	env["PWD"      ]  = scriptDirPath
	env["USER"     ]  = accountName

	ret = True

	try:
		process = subprocess.Popen(
			[ filePath ] + args,
			preexec_fn = _demote(uid, gid),
			cwd =  scriptDirPath,
			env = env,
		)
		exitCode = process.wait()

		if exitCode != 0:
			if log:
				log.error("Script " + repr(filePath) + " terminated with exit code: " + str(exitCode))
			ret = False
		else:
			if log:
				log.notice("Done running " + repr(filePath) + ".")

	except Exception as ee:
		if log:
			log.error(ee)
		ret = False

	return ret
#











__version__ = "0.2020.4.11.2"


from .Token import Token
from .RegExBasedTableTokenizer import RegExBasedTableTokenizer, RegExBasedTokenizingTable
from .RegExBasedTokenizer import RegExBasedTokenizer


from .AbstractTokenPattern import AbstractTokenPattern
from .TokenPattern import TokenPattern
from .TokenPatternAlternatives import TokenPatternAlternatives
from .TokenPatternRepeat import TokenPatternRepeat
from .TokenPatternSequence import TokenPatternSequence



